"""
    This package contains objects that can be used to determine the
    requirements of various interfaces used in WPILib. 
    
    Generally, the python version of WPILib does not require that you inherit
    from any of these interfaces, but instead will allow you to use custom 
    objects as long as they have the same methods.
"""

from .accelerometer import *
from .controller import *
from .counterbase import *
from .generichid import *
from .namedsendable import *
from .pidoutput import *
from .pidsource import *
from .potentiometer import *
from .speedcontroller import *
